////////////////////////////////////////////////////////////////////////////////
// wav2add.c
// this program reads a .wav file and spits out the harmonics ... suitable for
// use in a Kawai K5000 ADD source ... if you are not an electronic musician 
// with a Kawai K5 or Kawai K5000 then you probably don't have any use for it.
//
// copyright 1999 chris dalton (except as noted below)
// cdalton@grendelnet.com, chris.dalton@us.pwcglobal.com
// may be distributed under the terms of the GNU General Public License
//
// this program should compile on any un*x or win32 system (in the case of 
// windoze as a console application). if you want to run it on a system with
// a non x86 processor you may need to make changes because of the byte order 
// (big endian, little endian, cleveland endians, etc).
//
// if you make any changes or ports please send them to me.

#include "stdio.h"
#include "math.h"

#define TRUE 1
#define FALSE 0
#define MAX 4096

#define WAVOUTFILENAME "out.wav"

typedef struct {
  char id[4];
  long size;

  short          format_tag;
  unsigned short channels;
  unsigned long  samples_per_sec;
  unsigned long  avg_bytes_per_sec;
  unsigned short block_align;
  unsigned short bits_per_sample;
} format_chunk; // 24 bytes

typedef struct {
  char id[4];
  long size;
} data_chunk; // 8 bytes

// doublecomplex type
typedef struct { double r, i; } doublecomplex;
inline double zabs(doublecomplex z) { return sqrt(z.r*z.r + z.i*z.i); }

inline int k5k_scale(double h) 
{ 
  int level = ceil( 127. + 8. * log(fabs(h)) / log(2.) ); 
  return (level > 0) ? level : 0;
}

int fourier(char *c_dat, short *s_dat, int first, int count);
int fft(doublecomplex *z1, doublecomplex *z2, int *n, int *inzee);
int fftstp(doublecomplex *zin, int *after, int *now, int *before, doublecomplex *zout);
int options(int argc, char *argv[]);
void punt(int err);
void help();
void dump_wave(double h[MAX], doublecomplex z[MAX], int count);
void dump_add(double h[MAX], double h_max);

static double pi = M_PI;

////////////////////////////////////////////////////////////////////////////////
// global variables for command line options

int o_verbose = FALSE;
int o_begin = 0;
int o_end = 0;
int o_length = 0;
int o_quant = 1;
int o_trunc = 64;
char o_ka1_file[256] = "";

////////////////////////////////////////////////////////////////////////////////

int main(int argc, char *argv[])
{
  char header[12];
  unsigned char *c_samples;  // for 8 bit
  short *s_samples;          // for 16 bit
  format_chunk fmt;
  data_chunk dat;
  FILE *fp;
  int bytes_read, sample_i, byte_i;
  int wav_arg;
  short s;

  // process command line options
  if (argc < 2) 
    punt(1);
  else
    wav_arg = options(argc, argv);

  // open the input (.wav) file
  fp = fopen(argv[wav_arg], "rb");
  if (!fp) {
    fprintf(stderr, "could not open input file\n");
    return 2;
  } 

  // get the name of the ka1 file 
  if (wav_arg+1 < argc) {
    strcpy(o_ka1_file, argv[wav_arg+1]);
  } 

  // read the header
  // the file starts "RIFFxxxxWAVE" ... xxxx is the file size minus 8
  fread(header, sizeof(header), 1, fp);
  if (strncmp(header, "RIFF", 4)) {
    fprintf(stderr, "this file doesn't look like a .wav\n");
    fclose(fp);
    return 3;
  }
  
  // read the format chunk (TBD: don't assume it's the first chunk)
  fread(&fmt, sizeof(fmt), 1, fp);
  if (strncmp(fmt.id, "fmt ", 4)) {
    fprintf(stderr, "this wave file doesn't start with a format chunk\n");
    fclose(fp);
    return 4;
  }

  // we found a valid format chunk ... say something about it
  if (o_verbose) {
    printf("format chunk ...\n");
    printf("chunk size  : %d\n", fmt.size);
    printf("format tag  : %d\n", fmt.format_tag);
    printf("channels    : %d\n", fmt.channels);
    printf("samples/sec : %d\n", fmt.samples_per_sec);
    printf("bytes/sec   : %d\n", fmt.avg_bytes_per_sec);
    printf("block align : %d\n", fmt.block_align);
    printf("bits/sample : %d\n", fmt.bits_per_sample);
  }

  // read the data chunk (TBD: don't assume it's the second chunk)
  fread(&dat, sizeof(dat), 1, fp);
  if (strncmp(dat.id, "data", 4)) {
    fprintf(stderr, "the data chunk isn't right after the format chunk ... bummer\n");
    fclose(fp); 
    return 5;
  }
  
  // we found a data chunk ... not much to say but how long it is
  if (o_verbose) {
    printf("data chunk ...\n");
    printf("chunk size  : %d\n", dat.size);
  }

  // set o_end
  if (!o_end) {
    if (o_length) {
      o_end = o_begin + o_length;
    } else {
      o_end = dat.size/fmt.block_align - 1;
      if (o_end >= MAX)
        o_end = MAX - 1;
    }
  }

  // if this wave is uncompressed, mono and 8 or 16 bit read and print all the data
  if (fmt.format_tag == 1 && fmt.channels == 1 
      && (fmt.bits_per_sample == 8 || fmt.bits_per_sample == 16)
      && (fmt.block_align == 1 || fmt.block_align == 2)) 
  {
    c_samples = (char *)malloc(dat.size);
    if (!c_samples) {
      fprintf(stderr, "could not allocate memory for samples\n");
      fclose(fp); 
      return 6;
    }
    memset(c_samples, 0, dat.size);
    bytes_read = fread(c_samples, sizeof(char), dat.size, fp);
    if (bytes_read != dat.size) {
      fprintf(stderr, "unexpected end of file\n");
      fprintf(stderr, "%d\n", bytes_read);
      fprintf(stderr, "%d\n", dat.size);
      fclose(fp); 
      return 7;
    }

    if (fmt.bits_per_sample == 8) {
      // dump the sample data
      for (sample_i = 0; sample_i < dat.size; sample_i++)
        if (o_verbose && sample_i >= o_begin && sample_i <= o_end)
          printf("%3d: %3d\n", sample_i, c_samples[sample_i]);

      // do it
      fourier(c_samples, NULL, o_begin, (o_end - o_begin) + 1);
    } else {
      // dump the sample data
      s_samples = (short *)c_samples;
      for (sample_i = 0; sample_i < dat.size/2; sample_i++)
        if (o_verbose && sample_i >= o_begin && sample_i <= o_end)
          printf("%3d: %7d\n", sample_i, s_samples[sample_i]);

      // do it
      fourier(NULL, s_samples, o_begin, (o_end - o_begin) + 1);
    }

    free(c_samples);
  } else {
    fprintf(stderr, "wav file type not supported\n");
  }

  fclose(fp);

  return 0;
}

////////////////////////////////////////////////////////////////////////////////
// process command line options
// pre: argc >= 2 
// returns the index of the input wav argument

int options(int argc, char *argv[])
{
  int arg_i;
  char arg[256];

  for (arg_i=1; arg_i<argc; arg_i++) {
    // must be at least two characters
    if (strlen(argv[arg_i]) < 2) punt(2);

    // when we get to something without a dash we're done
    if (argv[arg_i][0] != '-')
      break;

    // check the arg against the list
    strcpy(arg, argv[arg_i]+1);
    if (arg[0] == 'v') {
      o_verbose = TRUE;
    } else if (arg[0] == 'h') {
      help();
    } else if (arg[0] == 'b') {
      if (strlen(arg) < 2) punt(4);
      o_begin = atoi(arg+1);
    } else if (arg[0] == 'e') { 
      if (strlen(arg) < 2) punt(5);
      o_end = atoi(arg+1);
    } else if (arg[0] == 'q') {
      if (strlen(arg) < 2) punt(6);
      o_quant = atoi(arg+1); 
    } else if (arg[0] == 't') {
      if (strlen(arg) < 2) punt(7);
      o_trunc = atoi(arg+1);
    } else if (arg[0] == 'l') {
      if (strlen(arg) < 2) punt(8);
      o_length = atoi(arg+1);
    } else {
      // none of the above
      fprintf(stderr, "%s\n", arg);             
      punt(20);
    }
  }

  return arg_i;
}

////////////////////////////////////////////////////////////////////////////////

static char usage[] = "usage: wav2add [options] foo.wav [bar.ka1]\n";

void punt(int err)
{
  fprintf(stderr, usage);
  fprintf(stderr, "wav2add -h for a help screen\n");
  fprintf(stderr, "code %d\n", err);
  exit(err);
}

////////////////////////////////////////////////////////////////////////////////

void help()
{
  fprintf(stderr, "wav2add version 0.1\n");
  fprintf(stderr, usage);
  fprintf(stderr, "-h:  print this help\n");
  fprintf(stderr, "-bN: sample point to begin with (default 0)\n");
  fprintf(stderr, "-eN: sample point to end with (default begin+100)\n");
  fprintf(stderr, "-tN: truncate at N harmonics (default 64)\n");
  fprintf(stderr, "-v:  be verbose (print harmonics and all sample data)\n");
  exit(1);
}

////////////////////////////////////////////////////////////////////////////////

int fourier(char *c_dat, short *s_dat, int first, int count)
{
  doublecomplex z1[MAX], z2[MAX]; // TBD: dynamic
  doublecomplex *zout;
  double h[MAX], h_max, sum;
  int i, j, inzee;
  int n;

  // set up the arrays
  // o_quant is how many sample points to average ... if o_quant doesn't divide 
  // count evenly we end up ignoring a few points at the end
  if (o_verbose)
    printf("function:\n");
  for (i = 0; i < count/o_quant; i++) {
    sum = 0.;
    for (j = 0; j < o_quant; j++) {
      if (c_dat)
        sum += (double)c_dat[first + i*o_quant + j];
      else
        sum += (double)s_dat[first + i*o_quant + j];
    }
    z1[i].r = sum;
    z1[i].i = 0.;
    z2[i].r = 0.;
    z2[i].i = 0.;
    if (o_verbose)
      printf("%14.6lf\n", z1[i].r);
//    printf("%3d: %14.6lf\n", i, z1[i].r);
//    printf("%3d: %14.6lf %14.6lf\n", i, z1[i].r, z1[i].i);
  }

  // call fft to do the transform
  n = count/o_quant;
  inzee = 1;
  fft(z1, z2, &n, &inzee);
  zout = (inzee == 1) ? z1 : z2;

  // calculate the harmonics, which are just the radii of the fourier coeffs, and 
  // find the max harmonic ... TBD: warn if it isn't #1
  h_max = 0.;
  for (i = 0; i<count/o_quant; i++) {
    h[i] = zabs(zout[i]);
    if (h[i] > h_max)
      h_max = h[i];
  }

  // dump the harmonics
  printf("harmonics:\n");
  for (i = 1; i<=o_trunc; i++) {
    printf("%5d:%3d", i, k5k_scale(h[i]/h_max));
    if (!(i%8))
      printf("\n");
  }
  if (o_trunc%8)
    printf("\n");

/*
  printf("harmonics:\n");
  printf("  #        raw          normalized    K5k \n");
  for (i = 0; i<=o_trunc; i++)
    printf("%3d: %14.6lf %14.6lf %6d\n", i, h[i], h[i]/h_max, k5k_scale(h[i]/h_max));
*/

  // dump a wav file for comparison
  dump_wave(h, zout, count);

  // put the harmonics in a ka1 file 
  dump_add(h, h_max);

  return 0;
}

////////////////////////////////////////////////////////////////////////////////
// put the harmonics in a ka1 file
// right now they go in the soft section of the first source ... if the file isn't
// a two-source ka1 with both sources ADD, the file will probably get hosed
// TBD: put the output in any source

void dump_add(double h[MAX], double h_max)
{
  FILE *fp;
  int i;
  unsigned char c, add_checksum;

  if (!strlen(o_ka1_file))
    return;

  fp = fopen(o_ka1_file, "r+");
  if (!fp)
    return;
  fseek(fp, 0x123, SEEK_SET); // this seems to be the place  

  if (o_trunc > 64) 
    fprintf(stderr, "only the first 64 harmonics will be written to the ka1 file\n");

  // overwrite the harmonic data in the soft section of the first source
  for (i = 1; i<=o_trunc; i++) {
     c = (unsigned char) k5k_scale(h[i]/h_max);
     fwrite(&c, sizeof(char), 1, fp);
  }

  // now that we've done the deed, we have to fix the checksum
  // calculate the check sum for "add wave kit" data
  fseek(fp, 0xFF, SEEK_SET); 
  add_checksum = 0;
  for (i = 0; i<805; i++) {
    fread(&c, sizeof(char), 1, fp);
    add_checksum += c;
  }
  add_checksum = (add_checksum + 0xA5) & 0x7F;
  fseek(fp, 0xFE, SEEK_SET);
  fwrite(&add_checksum, sizeof(char), 1, fp);

  fclose(fp);
}

////////////////////////////////////////////////////////////////////////////////
// dump a wav file from the harmonics
// output is always 44.1 Hz, 16 bit (so the pitch may change if the input is something else)

void dump_wave(double h[MAX], doublecomplex z[MAX], int count)
{
  format_chunk fmt;
  data_chunk dat;
  double f, sum_h, theta;
  int i, j;
  unsigned long len;
  short int f_short;
  FILE *fp;

  // open the output wav file
  fp = fopen(WAVOUTFILENAME, "wb");  
  if (!fp)
    return;

  // write the header, the format chunk, and the begining of the data chunk
  fwrite("RIFF", 4, 1, fp);
  len = sizeof(fmt) + sizeof(dat) + 4 + count*sizeof(f_short);
  fwrite(&len, sizeof(len), 1, fp);
  fwrite("WAVE", 4, 1, fp);
  memcpy(fmt.id, "fmt ", 4);
  fmt.size = sizeof(fmt) - 8;
  fmt.format_tag = 1;
  fmt.channels = 1;
  fmt.samples_per_sec = 44100;
  fmt.avg_bytes_per_sec = 88200;
  fmt.block_align = 2;
  fmt.bits_per_sample = 16;
  fwrite(&fmt, sizeof(fmt), 1, fp);
  memcpy(dat.id, "data", 4);
  dat.size = count*sizeof(f_short);
  fwrite(&dat, sizeof(dat), 1, fp);

  // sum the harmonics (so we have an upper bound on the amplitude)
  sum_h = 0.;
  for (j = 1; j<=o_trunc; j++)
    sum_h += h[j];

  // evaluate the "function"
  for (i = 0; i<count; i++) {
    theta = (2. * pi * i) / (double)count;

    // evaluate the series, ignoring coeffs above o_trunc
    f = 0.;
    for (j = 1; j<=o_trunc; j++)
//    f += z[j].r * cos(j*theta) - z[j].i * sin(j*theta); 	// real part of sum exp(i*n*theta)
      f += h[j] * sin(j*theta); 				// sine series, h = abs(z)

    f_short = (short int)(32767*f/sum_h);
    fwrite(&f_short, sizeof(f_short), 1, fp);
  }

  fclose(fp);
} 

/* ************************************************************ */
/* FROM HERE ON IS TRANSLATED FORTRAN                           */
/* ORIGINAL CREDIT:                                             */
/*    THIS CODE IS FEATURED IN ELEMENTARY NUMERICAL ANALYSIS,   */
/*    AN ALGORITHMIC APPROACH, THIRD ED., BY COUTE AND DEBOOR,  */
/*    MCGRAW-HILL, INC., 1980, PG. 281.                         */
/* ************************************************************ */

/* ******************************************************** */
/* CONSTRUCTS THE DISCRETE FOURIER TRANSFORM OF  Z1         */
/* (OR Z2) IN THE COOLEY-TUKEY WAY, BUT WITH A TWIST.       */
/* N IS DIVIDED INTO ITS PRIME FACTORS (TO A POINT) FOR     */
/* EACH FACTOR P, THE P-TRANSFORM OF APPR. P-SUBVECTORS     */
/* IS CALCULATED IN FFTSTP AND STORED IN Z2 (OR Z1).        */
/* ******************************************************** */

int fft(doublecomplex *z1, doublecomplex *z2, int *n, int *inzee)
{
    static int nextmx = 12;
    static int prime[12] = { 2,3,5,7,11,13,17,19,23,29,31,37 };

    int next, after, before;
    int now;

    /* Parameter adjustments */
    --z2;
    --z1;

    /* Function Body */
    after = 1;
    before = *n;
    next = 1;

L10:
    if (before / prime[next - 1] * prime[next - 1] < before) {
        ++next;
        if (next <= nextmx) {
            goto L10;
        } else {
            now = before;
            before = 1;
        }
    } else {
        now = prime[next - 1];
        before /= prime[next - 1];
    }

/* DEPENDENT ON INZEE CALL FFTSTP TO EVAL */
    if (*inzee == 1) {
        fftstp(&z1[1], &after, &now, &before, &z2[1]);
    } else {
        fftstp(&z2[1], &after, &now, &before, &z1[1]);
    }
    *inzee = 3 - *inzee;

    if (before == 1) {
        return 0;
    }
    after *= now;
    goto L10;
}

/* ***************************************************** */
/* FFTSTP CARRIES OUT ONE STEP OF THE DISCRETE FAST      */
/* FOURIER TRANSFORMATION.                               */
/* ***************************************************** */

int fftstp(doublecomplex *zin, int *after, int *now, int *before, doublecomplex *zout)
{
    static double twopi = 6.2831853071795864769;

    /* System generated locals */
    int zin_dim1, zin_dim2, zin_offset, zout_dim1, zout_dim2, zout_offset,
             i__2, i__3, i__4;
    double d__1, d__2;
    doublecomplex z__1, z__2;

    /* Local variables */
    int j;
    double angle;
    doublecomplex omega, value;
    int ia, ib, in;
    doublecomplex arg;

    /* Parameter adjustments */
    zout_dim1 = *after;
    zout_dim2 = *now;
    zout_offset = zout_dim1 * (zout_dim2 + 1) + 1;
    zout -= zout_offset;
    zin_dim1 = *after;
    zin_dim2 = *before;
    zin_offset = zin_dim1 * (zin_dim2 + 1) + 1;
    zin -= zin_offset;

    /* Function Body */
    angle = twopi / (double) (*now * *after);
    d__1 = cos(angle);
    d__2 = -sin(angle);
    z__1.r = d__1, z__1.i = d__2;
    omega.r = z__1.r, omega.i = z__1.i;
    arg.r = 1., arg.i = 0.;

/* PERFORM THE FOURIER TRANSFORM */
    for (j = 1; j <= *now; ++j) {
        i__2 = *after;
        for (ia = 1; ia <= i__2; ++ia) {
            i__3 = *before;
            for (ib = 1; ib <= i__3; ++ib) {
                i__4 = ia + (ib + *now * zin_dim2) * zin_dim1;
                value.r = zin[i__4].r, value.i = zin[i__4].i;
                for (in = *now - 1; in >= 1; --in) {
                    z__2.r = value.r * arg.r - value.i * arg.i, z__2.i = value.r * arg.i + value.i * arg.r;
                    i__4 = ia + (ib + in * zin_dim2) * zin_dim1;
                    z__1.r = z__2.r + zin[i__4].r, z__1.i = z__2.i + zin[i__4].i;
                    value.r = z__1.r, value.i = z__1.i;
                }
                i__4 = ia + (j + ib * zout_dim2) * zout_dim1;
                zout[i__4].r = value.r, zout[i__4].i = value.i;
            }
            z__1.r = arg.r * omega.r - arg.i * omega.i, z__1.i = arg.r * omega.i + arg.i * omega.r;
            arg.r = z__1.r, arg.i = z__1.i;
        }
    }

    return 0;
}
